/*************************************************
*    The PMS Music Typesetter - 2nd incarnation  *
*************************************************/

/* Copyright (c) Philip Hazel, 1991 - 2004 */

/* Written by Philip Hazel, starting November 1991 */
/* This file last modified: February 2004 */


/* This file contains code for reading and processing strings. */


#include "pmwhdr.h"
#include "readhdr.h"


#define string_start_size 64


static BOOL in_check_string = FALSE;

static uschar *music_escapes = US"bsmcQq#$%><udlr";

static uschar music_escape_values[] = {
  49, 50, 51, 53, 55, 57, 37, 39, 40, 122, 121, 126, 124, 123, 125 };



/*************************************************
*           Read a string of any length          *
*************************************************/

/* The string may extend over more than one line; newlines count as spaces. */

uschar *string_read(void)
{
int p = 0;
int size = string_start_size;
uschar *s;

sigch();
if (read_ch != '\"')
  {
  error_moan(10, "String in quotes");
  return NULL;
  }

s = store_Xget(size);
next_ch();

while (read_ch != '\"' && read_ch != EOF)
  {
  s[p++] = (read_ch == '\n')? ' ' : read_ch;
  if (p == size-1)
    {
    int increment = (size > 1024)? 1024 : size;
    uschar *ss = store_Xget(size + increment);
    memcpy(ss, s, size);
    store_free(s);
    s = ss;
    size += increment;
    }
  next_ch();
  }
s[p] = 0;
if (read_ch == EOF) { error_moan(23); return NULL; }
next_ch();
return s;
}



/*************************************************
*   Check the escapes in a string & transpose    *
*************************************************/

/* This is called for all but PostScript strings on reading, so that any errors
are given at that time. We set a flag while in it to ensure that the \a^ -> \a@
fudge works correctly.

A second job of this function is to look for note letter transpositions in the 
string and to carry them at at this time. Other escapes are re-processed later, 
at output time.

Argument:   the string
Returns:    the string, may be modified
*/

uschar *string_check(uschar *s)
{
int c;
uschar *ss, *tt;

/* If there are any transposed note names in the string, we must build a new
string with the transpositions done. */

if ((tt = Ustrstr(s, "\\t")) != NULL)
  {
  int p = tt - s; 
  int size = Ustrlen(s) * 2;
  ss = store_Xget(size); 
  
  memcpy(ss, s, p);
  
  for (s = tt; *s != 0; s++)
    {  
    int i, pitch, abspitch; 
    int note, acc;

    if (p >= size-9)
      {
      int increment = (size > 1024)? 1024 : size;
      uschar *sss = store_Xget(size + increment);
      memcpy(sss, ss, size);
      store_free(ss);
      ss = sss;
      size += increment;
      }
 
    if (*s != '\\' || s[1] != 't' || s[2] < 'A' || s[2] > 'G')
      {
      ss[p++] = *s;
      continue; 
      }  
       
    note = s[2]; 
    s += 2; 
    acc = ac_none;
    
    if (s[1] == '#') acc = ac_sharp;
    else if (s[1] == '$') acc = ac_flat;
    else if (s[1] == '%') acc = ac_natural;
    
    if (acc != ac_none) s++;
    
    pitch = 36 + read_basicpitch[note - 'A'];
    abspitch = pitch + read_accpitch[acc] - 2;
    
    /* If we are not in a stave, the appropriate values for the transposing
    function should have been fudged up by the heading reading code. */
    
    if (stave_transpose < max_transpose)
      abspitch = transpose_note(
        abspitch,    /* input absolute pitch */
        &pitch,      /* for output new pitch */
        &acc,        /* for output accidental */
        0,           /* do not force an output accidental */
        TRUE,        /* force addidental - not relevant for text */
        TRUE,        /* single note - not relevant for text */
        TRUE,        /* texttranspose */
        0);          /* tie count - not relevant for text */

    /* Retrieve a pitch within an octave */
    
    pitch %= 12;
    for (i = 0; i < 6; i++)
      {
      if (pitch == read_basicpitch[i]) break; 
      }  
    ss[p++] = 'A' + i;
    
    /* Handle accidentals. Double sharps and double flats are never seen, so 
    change to the appropriate enharmonic note. We should only ever get 
    "sensible" double accidentals. */
    
    switch (acc)
      {
      case ac_sharp:   Ustrcpy(ss+p, "\\*u*#*d*<\\"); p += 10; break; 
      case ac_flat:    Ustrcpy(ss+p, "\\*$*<\\"); p += 6; break; 

      #ifdef NEVER
      /* This code retains a natural if that note is sharpened or flattened in 
      the new key signature, but in practice that isn't what is actually seen 
      in printed music, where naturals never occur in chord names. */
       
      case ac_natural: 
      knumacc = main_keysigtable[stave_key_tp];  /* Number of accidentals */
      if (abs(knumacc) <= 7)                     /* Something odd otherwise */
        {
        uschar *nlist = (knumacc > 0)? US"FCGDAEB" : US"BEADGCF";
        for (i = 0; i < abs(knumacc); i++)
          {
          if (ss[p-1] == nlist[i]) break;  
          }
        if (i < abs(knumacc)) 
          {
          Ustrcpy(ss+p, "\\*%*<\\");  
          p += 6; 
          } 
        }  
      break;
      #endif 
      
      case ac_dsharp:
      if (ss[p-1] == 'G') ss[p-1] = 'A'; else ss[p-1] += 1;
      break;
      
      case ac_dflat:
      if (ss[p-1] == 'A') ss[p-1] = 'G'; else ss[p-1] -= 1;
      break;      
      }
    } 
    
 ss[p] = 0;
 s = ss; 
 }
 
/* Now do an escape check on the (new) string, but don't make any changes
to it. This just gets any errors early. */ 

ss = s;                     /* The string to return */
in_check_string = TRUE;
while ((c = *s++) != 0)
  {
  if (c == '\\')
    {
    int nf, r;
    uschar xs[80];
    s = string_escape(s, xs, &nf, &r);
    }
  }
in_check_string = FALSE;

return ss;
}




/*************************************************
*            Get an integer in a string          *
*************************************************/

/* The first digit is pointed to */

static int string_integer(uschar **p)
{
uschar *s = *p;
int n = 0;
while (isdigit(*s)) n = (n * 10) + (*s++ - '0');
if (n < 32 || n > 255) { error_moan(53); n = 32; }
*p = s;
return n;
}



/*************************************************
*             Decode string escape               *
*************************************************/

/* On entry, the pointer is past the initial '\';
return it past the whole escape. The data returned is:

  In ss, a string to be printed, which can be empty;
  In font, a font to change to before printing, set to
    -1 for no change; a pseudo-font is used for small caps;
  In revert, a flag saying whether to revert to the old
    font after printing - FALSE for font change escapes
    and TRUE for special uschar escapes.
*/


uschar *string_escape(uschar *s, uschar *ss, int *font, int *revert)
{
uschar *escape_uschar = s - 1;
uschar *error_arg = NULL;
int error_number = 54;
int offset, xoffset;

*font = -1;
*ss = 0;
*revert = FALSE;

/* Deal with font changes */

if (strncmp(s, "rm\\", 3) == 0) { *font = font_rm; return s+3; }
if (strncmp(s, "it\\", 3) == 0) { *font = font_it; return s+3; }
if (strncmp(s, "bf\\", 3) == 0) { *font = font_bf; return s+3; }
if (strncmp(s, "bi\\", 3) == 0) { *font = font_bi; return s+3; }
if (strncmp(s, "sy\\", 3) == 0) { *font = font_sy; return s+3; }
if (strncmp(s, "mu\\", 3) == 0) { *font = font_mu; return s+3; }
if (strncmp(s, "mf\\", 3) == 0) { *font = font_mf; return s+3; }
if (strncmp(s, "sc\\", 3) == 0) { *font = font_sc; return s+3; }

if (*s == 'x' && isdigit(s[1]))
  {
  if (s[2] == '\\')
    {
    *font = font_xx + s[1] - '1';
    return s+3;
    }
  else if (isdigit(s[2]))
    {
    int n = 10*(s[1] - '0') + s[2] - '0';
    if (n > MaxExtraFont) error_moan(27, MaxExtraFont);
      else *font = font_xx + n - 1;
    return s+4;
    }
  }

/* If not a font change, deal with other escapes */

*revert = TRUE;

/* Deal with ligatures */

if      (strncmp(s, "fi",  2) == 0) { *ss++ = 158; s += 2; }
else if (strncmp(s, "fl",  2) == 0) { *ss++ = 159; s += 2; }
else if (strncmp(s, "ss",  2) == 0) { *ss++ = 223; s += 2; }
else if (strncmp(s, "---", 3) == 0) { *ss++ = 153; s += 3; }
else if (strncmp(s, "--",  2) == 0) { *ss++ = 152; s += 2; }

else switch (*s)
  {
  case 0:         /* Backslash as last uschar is ignored */
  break;

  case '@':       /* starts within-string comment */
  while (*(++s) != '\\' && *s != 0);
  if (*s == '\\') s++;
  break;

  case '\\':
  *ss++ = *s++;   /* \\ gives \ */
  break;

  case '\'':      /* \' gives ascii quote */
  case '`':       /* \` gives ascii grave */
  *ss++ = *s;
  s++;
  break;

  case '*':       /* \* escapes into the music font */
  if (*(++s) == '*') { *font = font_mf; s++; }
    else *font = font_mu;

  /* Loop for multiple musical escapes */

  for (;;)
    {
    uschar *p;
    if ((p = Ustrchr(music_escapes, *s)) != NULL)
      {
      int n = p - music_escapes;
      *ss++ = music_escape_values[n];
      if (*(++s) == '.' && n <= 4)     /* Allow '.' after bsmcq */
        {
        *ss++ = 63;
        s++;
        }
      }
    else
      if (isdigit(*s)) *ss++ = string_integer(&s);
    else
      {
      error_arg = US"after \\* or \\** a music character code or number is";
      s++;
      }

    if (*s == '\\') { s++; break; }

    if (*s++ != '*')
      {
      error_arg = US"\"\\\" or \"*\"";
      break;
      }

    if (*font == font_mf && *s++ != '*')
      {
      error_arg = US"\"*\"";
      break;
      }
    }
  break;

  /* Non-musical escapes */

  case '?': *ss++ = 191; s++; break;  /* Spanish question mark */

  case 'a': offset = 224; xoffset = 1; goto ACCENT;
  case 'e': offset = 232; xoffset = 0; goto ACCENT;
  case 'i': offset = 236; xoffset = 0; goto ACCENT;
  case 'o': offset = 242; xoffset = 1; goto ACCENT;
  case 'u': offset = 249; xoffset = 0; goto ACCENT;
  case 'A': offset = 192; xoffset = 1; goto ACCENT;
  case 'E': offset = 200; xoffset = 0; goto ACCENT;
  case 'I': offset = 204; xoffset = 0; goto ACCENT;
  case 'O': offset = 210; xoffset = 1; goto ACCENT;
  case 'U': offset = 217; xoffset = 0; goto ACCENT;

ACCENT: s++;

  /* There is a problem with escapes using circumflex, because this is
  also used to split up underlay strings for centering. This is dealt with
  as follows: Since all strings are checked as they are read, if we are in
  the string_check function, we convert the circumflex into an at. During
  the display state, we recognize @. */

  switch(*s++)
    {
    case '`':  *ss++ = offset;   break;
    case '\'': *ss++ = offset+1; break;
    case '.':  *ss++ = offset+3+xoffset; break;

    case '^':          /* Should never see except while checking first time */
    *ss++ = offset+2;
    s[-1] = '@';
    break;

    case '@':
    if (!in_check_string)
      {
      *ss++ = offset+2;
      break;
      }
    /* Else fall through to the error */

    default:
    error_arg = US"\".\", \"^\", \"\'\", or \"`\"";
    s--;
    break;
    }
  break;


  case 'c': offset = 231; goto CEDILLA;
  case 'C': offset = 199; goto CEDILLA;

CEDILLA:

  if (*(++s) == ',') *ss++ = offset;
    else if (*s == ')') { *ss++ = 169; }
    else if (*s == ']') { *font = font_sy; *ss++ = 211; }
  else error_arg = US"\"\\c,\" or \"\\c)\"";
  s++;
  break;


  case 'n': offset = 241; goto TILDE;
  case 'N': offset = 209; goto TILDE;

TILDE:

  if (*(++s) == '~') *ss++ = offset;
  else error_arg = US"\"\\n~\"";
  s++;
  break;


  case 'p':
  if (*(++s) == '\\')
    {
    ss += sprintf(CS ss, "%d", curpage->number);
    s++;
    }
  else if (*s == 'o' && s[1] == '\\')
    {
    if ((curpage->number & 1) != 0) ss += sprintf(CS ss, "%d", curpage->number);
    s += 2;
    }
  else if (*s == 'e' && s[1] == '\\')
    {
    if ((curpage->number & 1) == 0) ss += sprintf(CS ss, "%d", curpage->number);
    s += 2;
    }
  else error_arg = US"\\p\\, \\po\\, or \\pe\\";
  break;


  case 's':
  if (strncmp(s+1, "e\\", 2) == 0)
    {
    s += 3;
    if ((curpage->number & 1) == 0) for (;;)
      {
      while (*s != '\\' && *s != 0) s++;
      if (*s == 0) break;
      if (strncmp(++s, "e\\", 2) == 0)
        {
        s += 2;
        break;
        }
      }
    }
  else if (strncmp(s+1, "o\\", 2) == 0)
    {
    s += 3;
    if ((curpage->number & 1) != 0) for (;;)
      {
      while (*s != '\\' && *s != 0) s++;
      if (*s == 0) break;
      if (strncmp(++s, "o\\", 2) == 0)
        {
        s += 2;
        break;
        }
      }
    }
  else if (isdigit(*(++s)))
    {
    *font = font_sy;
    *ss++ = string_integer(&s);
    if (*s++ != '\\')
      {
      error_number = 54;
      error_arg = US"\\";
      }
    }
  else error_arg = US"\\so\\ or \\se\\ or number";
  break;


  default:
  if (isdigit(*s))
    {
    *ss++ = string_integer(&s);
    if (*s++ != '\\') error_arg = US"\"\\\"";
    }
  else
    {
    error_number = 40;
    error_arg = (uschar *)(int)(*s);
    }
  break;
  }


/* If there has been an error, output the message with an
appropriate correction so the point is at the right place
for strings that are all on one line. Then skip to the end
of the string. Errors should only occur during the reading
phase when this routine is called for checking. Flatten the
original '\' character, to prevent a second error, which
otherwise happens when an underlay string is split. Also
and subsequent ones.*/

if (error_arg != NULL)
  {
  *escape_uschar = '/';
  while (*s) { *s++ = '/'; error_ptr_correction++; }
  error_ptr_correction += 1;
  error_moan(error_number, error_arg);
  }


/* Terminate the output string and return */

*ss = 0;
return s;
}




/*************************************************
*     Read string in stave and handle options    *
*************************************************/

void string_stavestring(BOOL rehearse)
{
b_textstr *p;
BOOL hadab = FALSE;
int rotate = 0;
int halfway = 0;
int htype = 0;
int size = -1;
int flags = stave_textflags;
int adjustx = 0;
int adjusty = 0;
int fontid;
uschar *s = string_read();

if (s == NULL) return;

/* Note that two successive slashes are a caesura, not
a bad string option. */

while (read_ch == '/' && *read_chptr != '/')
  {
  next_ch();
  switch(read_ch)
    {
    case 'a':
    hadab = TRUE;
    flags &= ~(text_ul|text_fb|text_ps|text_middle|text_atulevel|text_absolute);
    flags |= text_above;
    next_ch();
    if (read_ch == 'o')
      {
      flags |= text_atulevel;
      next_ch();
      }
    else if (isdigit(read_ch))
      {
      flags |= text_absolute;
      adjusty = read_integer(TRUE);
      }
    break;

    case 'b':
    next_ch();
    if (read_ch == 'o' && *read_chptr == 'x')
      {
      flags |= text_box;
      next_ch();
      next_ch();
      }
    else if (read_ch == 'a' && *read_chptr == 'r')
      {
      flags |= text_baralign;
      next_ch();
      next_ch();
      }
    else   /* /b or /bu */
      {
      hadab = TRUE;
      flags &= ~(text_ul | text_fb | text_above | text_ps |
        text_middle | text_atulevel | text_absolute);
      if (read_ch == 'u')
        {
        flags |= text_atulevel;
        next_ch();
        }
      else if (isdigit(read_ch))
        {
        flags |= text_absolute;
        adjusty = -read_integer(TRUE);
        }
      }
    break;

    case 'c':
    flags &= ~text_endalign;
    flags |= text_centre;
    next_ch();
    break;

    case 'd':
    adjusty -= read_movevalue();
    break;

    case 'e':
    flags &= ~text_centre;
    flags |= text_endalign;
    next_ch();
    break;

    case 'f':
    next_ch();
    if (read_ch == 'b')
      {
      flags &= ~(text_ul | text_above | text_ps | text_middle | text_atulevel);
      flags |= text_fb;
      }
    else error_moan(37, "/fb");
    next_ch();
    break;
    
    case 'h':
    next_ch();
    halfway = 500;
    if (isdigit(read_ch)) halfway = read_integer(TRUE);
    break;  

    case 'l':
    adjustx -= read_movevalue();
    break;

    case 'm':
    flags &= ~(text_ul|text_fb|text_above|text_ps|text_atulevel|text_absolute);
    flags |= text_middle;
    next_ch();
    break;
    
    case 'n':
    next_ch();
    if (read_ch == 'c') flags &= ~text_centre;
    else if (read_ch == 'e') flags &= ~text_endalign;
    else error_moan(37, "/nc or /ne"); 
    next_ch(); 
    break; 

    case 'o':
    next_ch();
    if (read_ch == 'l')
      {
      flags &= ~(text_fb | text_ps);
      flags |= text_ul | text_above;
      }
    else error_moan(37, "/ol");
    next_ch();
    break;

    case 'p':
    next_ch();
    if (read_ch == 's')
      {
      flags &= ~(text_above | text_fb | text_ul | text_middle |
        text_atulevel | text_absolute);
      flags |= text_ps;
      }
    else error_moan(37, "/ps");
    next_ch();
    break;

    case 'r':
    if (read_chptr[0] == 'i' && read_chptr[1] == 'n' && read_chptr[2] == 'g')
      {
      flags |= text_ring;
      next_ch();
      next_ch();
      next_ch();
      next_ch();
      }
    else if (read_chptr[0] == 'o' && read_chptr[1] == 't')
      {
      next_ch();
      next_ch();
      next_ch();
      read_expect_integer(&rotate, TRUE, TRUE);
      }
    else adjustx += read_movevalue();
    break;

    case 's':
    next_ch();
    if (read_expect_integer(&size, FALSE, FALSE))
      {
      if (--size < 0 || size >= MaxTextFont)
        { error_moan(39, MaxTextFont); size = 0; }
      }
    break;

    case 't':
    next_ch();
    if (read_ch == 's') flags |= text_timealign;
      else error_moan(37, "/ts");
    next_ch();
    break;

    case 'u':
    if (isdigit(*read_chptr)) adjusty += read_movevalue(); else
      {
      next_ch();
      if (read_ch == 'l')
        {
        flags &= ~(text_fb | text_above | text_ps);
        flags |= text_ul;
        }
      else error_moan(37, "/u<number> or /ul");
      next_ch();
      }
    break;

    default:
    error_moan(37, "/a, /ao, /b, /bar, /box, /bu, /d, /e, /fb, "
      "/h, /l, /m, /ol, /ps, /r, /ring, /s, /u or /ul");
    next_ch();
    break;


    /* The second/third string stuff must come last. The only way to get
    out of a case within a loop is by GOTO. */

    case '\"':
      {
      int adjust = 0;
      int size1, size2;
      int *sizeptr = &size1;
      uschar *s1 = string_read();
      uschar *s2 = NULL;
      uschar *s3 = NULL;

      if (size < 0)
        size1 = ((flags & text_above) == 0)? stave_ulsize : stave_olsize;
      else size1 = size;

      size2 = size1;

      /* Don't bother if there's been some kind of error. Otherwise,
      split the string at '|' if there is one. The final part is stored
      as a "third" extra string. */

      if (s1 != NULL)
        {
        for (s3 = s1; *s3; s3++)
          {
          if (*s3 == '|')
            {
            *s3++ = 0;
            break;
            }
          }
        if (*s3 == 0) s3 = NULL;

        /* Now read options, and possibly third string */

        while (read_ch == '/' && *read_chptr != '/')
          {
          next_ch();
          switch (read_ch)
            {
            case 'd':
            adjust -= read_movevalue();
            break;

            case 's':
            next_ch();
            if (read_expect_integer(sizeptr, FALSE, FALSE))
              {
              if ((*sizeptr -= 1) < 0 || *sizeptr >= MaxTextFont)
                { error_moan(39, MaxTextFont); *sizeptr = 0; }
              }
            break;

            case 'u':
            adjust += read_movevalue();
            break;

            case '\"':
            if (s2 != NULL) error_moan(92);
            s2 = string_read();
            sizeptr = &size2;
            break;

            default:
            error_moan(37, "/d, /s, /u, or string");
            next_ch();
            break;
            }
          }


        /* Only applies to {und,ov}erlay. Set up a hyphen-type block,
        or point to an identical one. */

        if ((flags & text_ul) == 0) error_moan(90); else
          {
          int font = ((flags & text_above) == 0)?
            stave_ulfont : stave_olfont;
          htypestr *h = main_htypes;
          htypestr **hh = &main_htypes;
          htype++;

          while (h != NULL)
            {
            if (strcmp(h->string1, s1) == 0 &&
               ((h->string2 == NULL && s2 == NULL) ||
                (h->string2 != NULL && s2 != NULL &&
                  strcmp(h->string2, s2) == 0)) &&
               ((h->string3 == NULL && s3 == NULL) ||
                (h->string3 != NULL && s3 != NULL &&
                  strcmp(h->string3, s3) == 0)) &&
               h->adjust == adjust && h->font == font &&
               h->size1 == size1 && h->size2 == size2) break;
            hh = &(h->next);
            h = *hh;
            htype++;
            }

          if (h == NULL)
            {
            h = store_Xget(sizeof(htypestr));
            *hh = h;
            h->next = NULL;
            h->string1 = s1;
            h->string2 = s2;
            h->string3 = s3;
            h->font = font;
            h->adjust = adjust;
            h->size1 = size1;
            h->size2 = size2;
            }
          }
        }
      }
    goto ENDSTRING;
    }

  sigch();
  }

ENDSTRING:

/* Unless this is a PostScript string, check its escapes */

if ((flags & text_ps) == 0) s = string_check(s);

/* If the absolute flag is set and we have not had /a or /b, then
add in the default absolute position. */

if ((flags & text_absolute) != 0 && !hadab)
  adjusty += stave_textabsolute;


/* If size unset, over-ride it for underlay, overlay, or figured bass */

if (size < 0)
  {
  if ((flags & text_ul) != 0)
    size = ((flags & text_above) == 0)? stave_ulsize : stave_olsize;
      else if ((flags & text_fb) != 0) size = stave_fbsize;
        else size = stave_textsize;
  }


/* Over-ride flags if requested for rehearsal strings, and sort out the
default font for the rest if necessary. */

if (rehearse)
  {
  flags = text_rehearse | text_above;
  fontid = curmovt->font_rehearse;
  }
else
  {
  fontid = ((flags & text_ul) != 0)?
    (((flags & text_above) == 0)? stave_ulfont : stave_olfont) :
      ((flags & text_fb) != 0)? stave_fbfont : stave_textfont;
  }


/* If this is a non-{und,ov}erlay string, create the text block
immediately, preceded by an extras block if required. */

if ((flags & text_ul) == 0)
  {
  if (rotate || halfway)
    {
    b_textXstr *pp = store_getitem(b_textX);
    pp->rotate = rotate;
    pp->halfway = halfway; 
    }
  p = store_getitem(b_text);
  p->size = size;
  p->font = fontid;
  p->flags = flags;
  p->string = s;
  p->x = adjustx;
  p->y = adjusty;
  }

/* Otherwise, create a block on the underlay chain to hold the string
until it is all parcelled out to the succeeding notes. Give an error
if rotation is attempted. */

else
  {
  BOOL overlay = (flags & text_above) != 0;
  int count = overlay? olay_offset : 0;
  ulaypend **pp = &stave_pendulay;
  ulaypend *p = stave_pendulay;
  ulaypend *q;

  if (rotate) error_moan(98, "supported", "for underlay or overlay");

  while (p != NULL)
    {
    if (p->level > count) break;
    if (overlay)
      {
      if (p->level >= olay_offset) count++;
      }
    else if (p->level < olay_offset) count++;
    pp = &(p->next);
    p = p->next;
    }

  q = store_Xget(sizeof(ulaypend));
  q->next = p;
  *pp = q;

  q->level = count;
  q->size = size;
  q->htype = htype;
  q->font = fontid;
  q->flags = flags;
  q->string = s;
  q->x = adjustx;
  q->y = adjusty;
  q->halfway = halfway; 
  }
}



/*************************************************
*              Get width of a PMW string         *
*************************************************/

/* This almost duplicates code in out_string, and yet it seems
hard to find a tidy way of combining them. The initial font may
have font_sc added for small caps. The final font is left in
string_font for use if needed. */

int string_width(uschar *s, int f, int pointsize)
{
int yield = 0;
int yield_height = 0;
int orig_pointsize = pointsize;
register int c;
uschar ss[256];
uschar *pp = ss;

DEBUG(("string_width() \"%s\"\n", s));

if (f >= font_sc)
  {
  pointsize = (pointsize * curmovt->smallcapsize) / 1000;
  f -= font_sc;
  }   

while ((c = *s++) != 0)
  {
  
  /* If the string buffer is getting full, measure the string so far. This
  leaves plenty of room for escapes (which in practice are only a few chars). */

  if (pp - ss > 250)
    {
    *pp = 0;
    yield += font_stringwidth(ss, f, pointsize);
    yield_height += font_stringheight; 
    pp = ss;
    }  

  /* Until we hit the escape character, '\', simply copy chars,
  except for quote and grave, which turn into close and open
  quotes (except in the music font). */

  if (c != '\\')
    {
    if (c == '\'' && f != font_mf && f != font_mu && f != font_sy)
      c = 145;
    else if (c == '`' && f != font_mf && f != font_mu && f != font_sy)
      c = 144;
    *pp++ = c;
    }

  else

  /* Interpret the escape. It may return a string to print and/or
  a font change, and a flag to say whether the font change is
  permanent or not */

    {
    int nf, r;
    uschar xs[80];

    s = string_escape(s, xs, &nf, &r);

    /* If there is a font change, first measure the string so far in the
    old font (if any). */

    if (nf >= 0)
      {
      if (pp != ss)
        {
        *pp = 0;
        yield += font_stringwidth(ss, f, pointsize);
        yield_height += font_stringheight; 
        pp = ss;
        }

      /* If the new font is temporary, measure the escaped string and
      mark it as empty. Otherwise, change the current font. */

      if (r)
        {
        yield += font_stringwidth(xs, nf, pointsize);
        yield_height += font_stringheight; 
        xs[0] = 0;
        }

      /* If the new font is "small caps", leave the font alone, but
      adjust the size. Otherwise reset the original size. */

      else 
        {
        if (nf == font_sc)
          {
          if (pointsize == orig_pointsize)
            pointsize = (pointsize * curmovt->smallcapsize) / 1000;  
          }
        else
          {     
          f = nf;
          pointsize = orig_pointsize;
          }  
        } 
      }

    /* Join the escape string onto the string so far */

    Ustrcpy(pp, xs);
    pp += Ustrlen(xs);
    }
  }

/* Catch any pending characters at the end */

if (pp != ss)
  {
  *pp = 0;
  yield += font_stringwidth(ss, f, pointsize);
  yield_height += font_stringheight; 
  }

string_font = f;                     /* stash away final font */
font_stringheight = yield_height;    /* put total height in known place */

DEBUG(("string_width() = %d\n", yield));
return yield;
}


/* End of string.c */
